"""The testing subpackage provides reasonable testing functionality for
building programmatic tests with and of geospatial data.  It provides functions
to generate input data of both raster and vector formats, and offers assertions
to verify the correctness of geospatial outputs.

Most useful features of this package have been exposed at the
pygeoprocessing.testing level.

Select locations have been chosen for their spatial references.  These
references are available in the ``sampledata`` module and al have the prefix
``SRS_``::

    from pygeoprocessing.testing import sampledata

    sampledata.SRS_WILLAMETTE
    sampledata.SRS_COLOMBIA


For writing tests, write them as you normally do!  Assertions are generally
file-based and raise ``AssertionError`` when a failure is encountered.

This example is relatively simplistic, since there will often be many more
assertions you may need to make to be able to test your model
effectively::

    import unittest
    import pygeoprocessing.testing
    import natcap.invest.example_model

    class ExampleTest(unittest.TestCase):
        def test_some_model(self):
            example_args = {
                'workspace_dir': './workspace',
                'arg_1': 'foo',
                'arg_2': 'bar',
            }
            natcap.invest.example_model.execute(example_args)

            # example assertion
            pygeoprocessing.testing.assert_rasters_equal('workspace/raster_1.tif',
                'regression_data/raster_1.tif')

"""
from __future__ import absolute_import


from .assertions import assert_close, isclose, \
    assert_rasters_equal, assert_vectors_equal, assert_csv_equal, \
    assert_md5_equal, assert_json_equal, assert_text_equal, \
    assert_checksums_equal
from .utils import digest_file, digest_file_list, digest_folder, \
    checksum_folder
from .sampledata import create_raster_on_disk, create_vector_on_disk

__all__ = [
    'create_raster_on_disk',
    'create_vector_on_disk',
    'digest_file',
    'digest_file_list',
    'digest_folder',
    'checksum_folder',
    'assert_close',
    'isclose',
    'assert_rasters_equal',
    'assert_vectors_equal',
    'assert_csv_equal',
    'assert_md5_equal',
    'assert_json_equal',
    'assert_text_equal',
    'assert_checksums_equal',
]



"""pygeoprocessing: geoprocessing routines for GIS.

__init__ module imports all the geoprocessing functions into this namespace.
"""
from __future__ import absolute_import

import types
import sys

import pkg_resources

from . import geoprocessing

from .geoprocessing_core import calculate_slope

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    # Package is not installed, so the package metadata is not available.
    # This should only happen if a package is importable but the package
    # metadata is not, as might happen if someone copied files into their
    # system site-packages or they're importing this package from the CWD.
    raise RuntimeError(
        "Could not load version from installed metadata.\n\n"
        "This is often because the package was not installed properly. "
        "Ensure that the package is installed in a way that the metadata is "
        "maintained.  Calls to ``pip`` and this package's ``setup.py`` "
        "maintain metadata.  Examples include:\n"
        "  * python setup.py install\n"
        "  * python setup.py develop\n"
        "  * pip install <distribution>")

__all__ = ('calculate_slope',)
for attrname in dir(geoprocessing):
    attribute = getattr(geoprocessing, attrname)
    if isinstance(attribute, types.FunctionType):
        __all__ += (attrname,)
        setattr(sys.modules['pygeoprocessing'], attrname, attribute)



"""
Fetch the subversion data repository based on the configuration
stored in svn_config.json.

To execute:
    $ python get_data.py
"""


import os
import json
import imp


# Import pygeoprocessing.testing.scm from source instead of via pygeoprocessing
# Handy, since I might not want to have to install pygeoprocessing to clone
# svn data.  Also, scm doesn't import pygeoprocessing, so that's cool.
scm = imp.load_source('scm', os.path.abspath(
    os.path.join(os.path.dirname(__file__),
                 '..', 'pygeoprocessing', 'testing', 'scm.py')))

_FILE = os.path.abspath(os.path.dirname(__file__))
SVN_CONFIG = os.path.join(_FILE, 'svn_config.json')

if __name__ == '__main__':
    config = json.load(open(SVN_CONFIG))
    data_dir = os.path.join(_FILE, config['local'])
    scm.checkout_svn(data_dir, config['remote'], config['rev'])




import os
import tempfile
import logging
import time
import sys
import traceback

cimport numpy
import numpy
cimport cython
from libcpp.map cimport map

from libc.math cimport sqrt
from libc.math cimport exp
from libc.math cimport ceil

from osgeo import gdal
import pygeoprocessing

DEFAULT_GTIFF_CREATION_OPTIONS = (
    'TILED=YES', 'BIGTIFF=YES', 'COMPRESS=LZW',
    'BLOCKXSIZE=256', 'BLOCKYSIZE=256')
LOGGER = logging.getLogger('pygeoprocessing.geoprocessing_core')

cdef float NODATA = -1.0


class MaskWrapper(object):
    def __init__(self, base_nodata):
        self.base_nodata = base_nodata

    def __call__(self, base_array):
        result = numpy.empty(base_array.shape, dtype=numpy.int8)
        result[:] = NODATA
        valid_mask = base_array != self.base_nodata
        result[valid_mask] = base_array[valid_mask] != 0
        return result

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def distance_transform_edt(base_mask_raster_path_band, target_distance_path):
    """Calculate the Euclidean distance transform.

    Parameters:
        base_mask_raster_path_band (tuple): a (path, band index) tuple to
            calculate value from non-zero valued pixels.

        target_distance_path (string): a path to a raster created by this
        function with same dimensions and projection as base_mask_path where
        all non-zero values of base_mask_path are equal to the euclidean
        distance to the closest 0 pixel.

    Returns:
        None."""
    cdef int yoff, row_index, block_ysize, win_ysize, n_rows
    cdef int xoff, block_xsize, win_xsize, n_cols
    cdef int q_index, local_x_index, local_y_index, u_index
    cdef int gu, gsq, tq, sq
    cdef numpy.ndarray[numpy.int32_t, ndim=2] g_block
    cdef numpy.ndarray[numpy.int32_t, ndim=1] s_array
    cdef numpy.ndarray[numpy.int32_t, ndim=1] t_array
    cdef numpy.ndarray[numpy.float32_t, ndim=2] dt
    cdef numpy.ndarray[numpy.int8_t, ndim=2] mask_block

    file_handle, base_mask_path = tempfile.mkstemp()
    os.close(file_handle)
    base_raster_info = pygeoprocessing.get_raster_info(
        base_mask_raster_path_band[0])
    base_nodata = base_raster_info['nodata'][
        base_mask_raster_path_band[1]-1]

    pygeoprocessing.raster_calculator(
        [base_mask_raster_path_band], MaskWrapper(base_nodata),
        base_mask_path,
        gdal.GDT_Byte, NODATA, calc_raster_stats=False)

    base_mask_raster = gdal.OpenEx(base_mask_path)
    base_mask_band = base_mask_raster.GetRasterBand(1)

    n_cols = base_mask_raster.RasterXSize
    n_rows = base_mask_raster.RasterYSize

    file_handle, g_path = tempfile.mkstemp()
    os.close(file_handle)
    raster_info = pygeoprocessing.get_raster_info(
        base_mask_raster_path_band[0])
    pygeoprocessing.new_raster_from_base(
        base_mask_raster_path_band[0], g_path, gdal.GDT_Int32, [NODATA],
        fill_value_list=None)
    g_raster = gdal.OpenEx(g_path, gdal.GA_Update)
    g_band = g_raster.GetRasterBand(1)
    g_band_blocksize = g_band.GetBlockSize()

    numerical_inf = (
        raster_info['raster_size'][0] + raster_info['raster_size'][1])
    # scan 1
    done = False
    block_xsize = raster_info['block_size'][0]
    mask_block = numpy.empty((n_rows, block_xsize), dtype=numpy.int8)
    g_block = numpy.empty((n_rows, block_xsize), dtype=numpy.int32)
    for xoff in numpy.arange(0, n_cols, block_xsize):
        win_xsize = block_xsize
        if xoff + win_xsize > n_cols:
            win_xsize = n_cols - xoff
            mask_block = numpy.empty((n_rows, win_xsize), dtype=numpy.int8)
            g_block = numpy.empty((n_rows, win_xsize), dtype=numpy.int32)
            done = True
        base_mask_band.ReadAsArray(
            xoff=xoff, yoff=0, win_xsize=win_xsize, win_ysize=n_rows,
            buf_obj=mask_block)
        # base case
        g_block[0, :] = (mask_block[0, :] == 0) * numerical_inf
        for row_index in range(1, n_rows):
            for local_x_index in range(win_xsize):
                if mask_block[row_index, local_x_index] == 1:
                    g_block[row_index, local_x_index] = 0
                else:
                    g_block[row_index, local_x_index] = (
                        g_block[row_index-1, local_x_index] + 1)
        for row_index in range(n_rows-2, -1, -1):
            for local_x_index in range(win_xsize):
                if (g_block[row_index+1, local_x_index] <
                        g_block[row_index, local_x_index]):
                    g_block[row_index, local_x_index] = (
                        1 + g_block[row_index+1, local_x_index])
        g_band.WriteArray(g_block, xoff=xoff, yoff=0)
        if done:
            break
    g_band.FlushCache()

    pygeoprocessing.new_raster_from_base(
        base_mask_raster_path_band[0], target_distance_path.encode('utf-8'),
        gdal.GDT_Float32, [NODATA], fill_value_list=None)
    target_distance_raster = gdal.OpenEx(target_distance_path, gdal.GA_Update)
    target_distance_band = target_distance_raster.GetRasterBand(1)

    LOGGER.info('Distance Transform Phase 2')

    s_array = numpy.empty(n_cols, dtype=numpy.int32)
    t_array = numpy.empty(n_cols, dtype=numpy.int32)

    done = False
    block_ysize = g_band_blocksize[1]
    g_block = numpy.empty((block_ysize, n_cols), dtype=numpy.int32)
    dt = numpy.empty((block_ysize, n_cols), dtype=numpy.float32)
    sq = 0  # initialize so compiler doesn't complain
    gsq = 0
    for yoff in numpy.arange(0, n_rows, block_ysize):
        win_ysize = block_ysize
        if yoff + win_ysize >= n_rows:
            win_ysize = n_rows - yoff
            g_block = numpy.empty((win_ysize, n_cols), dtype=numpy.int32)
            dt = numpy.empty((win_ysize, n_cols), dtype=numpy.float32)
            done = True
        g_band.ReadAsArray(
            xoff=0, yoff=yoff, win_xsize=n_cols, win_ysize=win_ysize,
            buf_obj=g_block)
        for local_y_index in range(win_ysize):
            q_index = 0
            s_array[0] = 0
            t_array[0] = 0
            for u_index in range(1, n_cols):
                gu = g_block[local_y_index, u_index]**2
                while (q_index >= 0):
                    tq = t_array[q_index]
                    sq = s_array[q_index]
                    gsq = g_block[local_y_index, sq]**2
                    if ((tq-sq)**2 + gsq <= (tq-u_index)**2 + gu):
                        break
                    q_index -= 1
                if q_index < 0:
                    q_index = 0
                    s_array[0] = u_index
                    sq = u_index
                    gsq = g_block[local_y_index, sq]**2
                else:
                    w = 1 + (
                        u_index**2 - sq**2 + gu - gsq) / (2*(u_index-sq))
                    if w < n_cols:
                        q_index += 1
                        s_array[q_index] = u_index
                        t_array[q_index] = w

            sq = s_array[q_index]
            gsq = g_block[local_y_index, sq]**2
            tq = t_array[q_index]
            for u_index in range(n_cols-1, -1, -1):
                dt[local_y_index, u_index] = (u_index-sq)**2+gsq
                if u_index == tq:
                    q_index -= 1
                    if q_index >= 0:
                        sq = s_array[q_index]
                        gsq = g_block[local_y_index, sq]**2
                        tq = t_array[q_index]

        valid_mask = g_block != NODATA
        dt[valid_mask] = numpy.sqrt(dt[valid_mask])
        dt[~valid_mask] = NODATA
        target_distance_band.WriteArray(dt, xoff=0, yoff=yoff)

        # we do this in the case where the blocksize is many times larger than
        # the raster size so we don't re-loop through the only block
        if done:
            break

    target_distance_band.FlushCache()
    gdal.Dataset.__swig_destroy__(target_distance_raster)
    gdal.Dataset.__swig_destroy__(base_mask_raster)
    gdal.Dataset.__swig_destroy__(g_raster)
    try:
        os.remove(g_path)
    except OSError:
        LOGGER.warning("couldn't remove file %s" % g_path)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def calculate_slope(
        base_elevation_raster_path_band, target_slope_path,
        gtiff_creation_options=DEFAULT_GTIFF_CREATION_OPTIONS):
    """Create a percent slope raster from DEM raster.

    Base algorithm is from Zevenbergen & Thorne "Quantitative Analysis of Land
    Surface Topography" 1987 although it has been modified to include the
    diagonal pixels by classic finite difference analysis.

    For the following notation, we define each pixel's DEM value by a letter
    with this spatial scheme:

        abc
        def
        ghi

    Then the slope at e is defined at ([dz/dx]^2 + [dz/dy]^2)^0.5

    Where

    [dz/dx] = ((c+2f+i)-(a+2d+g)/(8*x_cell_size)
    [dz/dy] = ((g+2h+i)-(a+2b+c))/(8*y_cell_size)

    In cases where a cell is nodata, we attempt to use the middle cell inline
    with the direction of differentiation (either in x or y direction).  If
    no inline pixel is defined, we use `e` and multiply the difference by
    2^0.5 to account for the diagonal projection.

    Parameters:
        base_elevation_raster_path_band (string): a path/band tuple to a
            raster of height values. (path_to_raster, band_index)
        target_slope_path (string): path to target slope raster; will be a
            32 bit float GeoTIFF of same size/projection as calculate slope
            with units of percent slope.
        gtiff_creation_options (list or tuple): list of strings that will be
            passed as GDAL "dataset" creation options to the GTIFF driver.

    Returns:
        None
    """
    cdef numpy.npy_float64 a, b, c, d, e, f, g, h, i, dem_nodata
    cdef numpy.npy_float64 x_cell_size, y_cell_size,
    cdef numpy.npy_float64 dzdx_accumulator, dzdy_accumulator
    cdef int row_index, col_index, n_rows, n_cols,
    cdef int x_denom_factor, y_denom_factor, win_xsize, win_ysize
    cdef numpy.ndarray[numpy.npy_float64, ndim=2] dem_array
    cdef numpy.ndarray[numpy.npy_float64, ndim=2] slope_array
    cdef numpy.ndarray[numpy.npy_float64, ndim=2] dzdx_array
    cdef numpy.ndarray[numpy.npy_float64, ndim=2] dzdy_array

    dem_raster = gdal.OpenEx(base_elevation_raster_path_band[0])
    dem_band = dem_raster.GetRasterBand(base_elevation_raster_path_band[1])
    dem_info = pygeoprocessing.get_raster_info(base_elevation_raster_path_band[0])
    dem_nodata = dem_info['nodata'][0]
    x_cell_size, y_cell_size = dem_info['pixel_size']
    n_cols, n_rows = dem_info['raster_size']
    cdef numpy.npy_float64 slope_nodata = numpy.finfo(numpy.float32).min
    pygeoprocessing.new_raster_from_base(
        base_elevation_raster_path_band[0], target_slope_path, gdal.GDT_Float32,
        [slope_nodata], fill_value_list=[float(slope_nodata)],
        gtiff_creation_options=gtiff_creation_options)
    target_slope_raster = gdal.OpenEx(target_slope_path, gdal.GA_Update)
    target_slope_band = target_slope_raster.GetRasterBand(1)

    for block_offset in pygeoprocessing.iterblocks(
            base_elevation_raster_path_band[0], offset_only=True):
        block_offset_copy = block_offset.copy()
        # try to expand the block around the edges if it fits
        x_start = 1
        win_xsize = block_offset['win_xsize']
        x_end = win_xsize+1
        y_start = 1
        win_ysize = block_offset['win_ysize']
        y_end = win_ysize+1

        if block_offset['xoff'] > 0:
            block_offset_copy['xoff'] -= 1
            block_offset_copy['win_xsize'] += 1
            x_start -= 1
        if block_offset['xoff']+win_xsize < n_cols:
            block_offset_copy['win_xsize'] += 1
            x_end += 1
        if block_offset['yoff'] > 0:
            block_offset_copy['yoff'] -= 1
            block_offset_copy['win_ysize'] += 1
            y_start -= 1
        if block_offset['yoff']+win_ysize < n_rows:
            block_offset_copy['win_ysize'] += 1
            y_end += 1

        dem_array = numpy.empty(
            (win_ysize+2, win_xsize+2),
            dtype=numpy.float64)
        dem_array[:] = dem_nodata
        slope_array = numpy.empty(
            (win_ysize, win_xsize),
            dtype=numpy.float64)
        dzdx_array = numpy.empty(
            (win_ysize, win_xsize),
            dtype=numpy.float64)
        dzdy_array = numpy.empty(
            (win_ysize, win_xsize),
            dtype=numpy.float64)

        dem_band.ReadAsArray(
            buf_obj=dem_array[y_start:y_end, x_start:x_end],
            **block_offset_copy)

        for row_index in range(1, win_ysize+1):
            for col_index in range(1, win_xsize+1):
                # Notation of the cell below comes from the algorithm
                # description, cells are arraged as follows:
                # abc
                # def
                # ghi
                e = dem_array[row_index, col_index]
                if e == dem_nodata:
                    # we use dzdx as a guard below, no need to set dzdy
                    dzdx_array[row_index-1, col_index-1] = slope_nodata
                    continue
                dzdx_accumulator = 0.0
                dzdy_accumulator = 0.0
                x_denom_factor = 0
                y_denom_factor = 0
                a = dem_array[row_index-1, col_index-1]
                b = dem_array[row_index-1, col_index]
                c = dem_array[row_index-1, col_index+1]
                d = dem_array[row_index, col_index-1]
                f = dem_array[row_index, col_index+1]
                g = dem_array[row_index+1, col_index-1]
                h = dem_array[row_index+1, col_index]
                i = dem_array[row_index+1, col_index+1]

                # a - c direction
                if a != dem_nodata and c != dem_nodata:
                    dzdx_accumulator += a - c
                    x_denom_factor += 2
                elif a != dem_nodata and b != dem_nodata:
                    dzdx_accumulator += a - b
                    x_denom_factor += 1
                elif b != dem_nodata and c != dem_nodata:
                    dzdx_accumulator += b - c
                    x_denom_factor += 1
                elif a != dem_nodata:
                    dzdx_accumulator += (a - e) * 2**0.5
                    x_denom_factor += 1
                elif c != dem_nodata:
                    dzdx_accumulator += (e - c) * 2**0.5
                    x_denom_factor += 1

                # d - f direction
                if d != dem_nodata and f != dem_nodata:
                    dzdx_accumulator += 2 * (d - f)
                    x_denom_factor += 4
                elif d != dem_nodata:
                    dzdx_accumulator += 2 * (d - e)
                    x_denom_factor += 2
                elif f != dem_nodata:
                    dzdx_accumulator += 2 * (e - f)
                    x_denom_factor += 2

                # g - i direction
                if g != dem_nodata and i != dem_nodata:
                    dzdx_accumulator += g - i
                    x_denom_factor += 2
                elif g != dem_nodata and h != dem_nodata:
                    dzdx_accumulator += g - h
                    x_denom_factor += 1
                elif h != dem_nodata and i != dem_nodata:
                    dzdx_accumulator += h - i
                    x_denom_factor += 1
                elif g != dem_nodata:
                    dzdx_accumulator += (g - e) * 2**0.5
                    x_denom_factor += 1
                elif i != dem_nodata:
                    dzdx_accumulator += (e - i) * 2**0.5
                    x_denom_factor += 1

                # a - g direction
                if a != dem_nodata and g != dem_nodata:
                    dzdy_accumulator += a - g
                    y_denom_factor += 2
                elif a != dem_nodata and d != dem_nodata:
                    dzdy_accumulator += a - d
                    y_denom_factor += 1
                elif d != dem_nodata and g != dem_nodata:
                    dzdy_accumulator += d - g
                    y_denom_factor += 1
                elif a != dem_nodata:
                    dzdy_accumulator += (a - e) * 2**0.5
                    y_denom_factor += 1
                elif g != dem_nodata:
                    dzdy_accumulator += (e - g) * 2**0.5
                    y_denom_factor += 1

                # b - h direction
                if b != dem_nodata and h != dem_nodata:
                    dzdy_accumulator += 2 * (b - h)
                    y_denom_factor += 4
                elif b != dem_nodata:
                    dzdy_accumulator += 2 * (b - e)
                    y_denom_factor += 2
                elif h != dem_nodata:
                    dzdy_accumulator += 2 * (e - h)
                    y_denom_factor += 2

                # c - i direction
                if c != dem_nodata and i != dem_nodata:
                    dzdy_accumulator += c - i
                    y_denom_factor += 2
                elif c != dem_nodata and f != dem_nodata:
                    dzdy_accumulator += c - f
                    y_denom_factor += 1
                elif f != dem_nodata and i != dem_nodata:
                    dzdy_accumulator += f - i
                    y_denom_factor += 1
                elif c != dem_nodata:
                    dzdy_accumulator += (c - e) * 2**0.5
                    y_denom_factor += 1
                elif i != dem_nodata:
                    dzdy_accumulator += (e - i) * 2**0.5
                    y_denom_factor += 1

                if x_denom_factor != 0:
                    dzdx_array[row_index-1, col_index-1] = (
                        dzdx_accumulator / (x_denom_factor * x_cell_size))
                else:
                    dzdx_array[row_index-1, col_index-1] = 0.0
                if y_denom_factor != 0:
                    dzdy_array[row_index-1, col_index-1] = (
                        dzdy_accumulator / (y_denom_factor * y_cell_size))
                else:
                    dzdy_array[row_index-1, col_index-1] = 0.0
        valid_mask = dzdx_array != slope_nodata
        slope_array[:] = slope_nodata
        # multiply by 100 for percent output
        slope_array[valid_mask] = 100.0 * numpy.sqrt(
            dzdx_array[valid_mask]**2 + dzdy_array[valid_mask]**2)
        target_slope_band.WriteArray(
            slope_array, xoff=block_offset['xoff'],
            yoff=block_offset['yoff'])

    dem_band = None
    target_slope_band = None
    gdal.Dataset.__swig_destroy__(dem_raster)
    gdal.Dataset.__swig_destroy__(target_slope_raster)
    dem_raster = None
    target_slope_raster = None


@cython.boundscheck(False)
def stats_worker(stats_work_queue, exception_queue):
    """Worker to calculate continuous min, max, mean and standard deviation.

    Parameters:
        stats_work_queue (Queue): a queue of 1D numpy arrays or None. If
            None, function puts a (min, max, mean, stddev) tuple to the
            queue and quits.

    Returns:
        None

    """
    cdef numpy.ndarray[numpy.float64_t, ndim=1] block
    cdef double M_local = 0.0
    cdef double S_local = 0.0
    cdef double min_value = 0.0
    cdef double max_value = 0.0
    cdef double x = 0.0
    cdef int i, n_elements
    cdef long long n = 0L
    payload = None

    try:
        while True:
            payload = stats_work_queue.get()
            if payload is None:
                LOGGER.debug('payload is None, terminating')
                break
            block = payload.astype(numpy.float64)
            n_elements = block.size
            with nogil:
                for i in range(n_elements):
                    n = n + 1
                    x = block[i]
                    if n <= 0:
                        with gil:
                            LOGGER.error('invalid value for n %s' % n)
                    if n == 1:
                        M_local = x
                        S_local = 0.0
                        min_value = x
                        max_value = x
                    else:
                        M_last = M_local
                        M_local = M_local+(x - M_local)/<double>(n)
                        S_local = S_local+(x-M_last)*(x-M_local)
                        if x < min_value:
                            min_value = x
                        elif x > max_value:
                            max_value = x

        if n > 0:
            stats_work_queue.put(
                (min_value, max_value, M_local, (S_local / <double>n) ** 0.5))
        else:
            LOGGER.warning(
                "No valid pixels were received, sending None.")
            stats_work_queue.put(None)
    except Exception as e:
        LOGGER.exception(
            "exception %s %s %s %s %s", x, M_local, S_local, n, payload)
        exception_queue.put(e)
        while not stats_work_queue.empty():
            stats_work_queue.get()
        raise



# coding=utf-8
"""Assertions for geospatial testing."""
from builtins import zip
from builtins import range

import os
import csv
import logging
import json
import itertools

from osgeo import gdal
from osgeo import ogr
from osgeo import osr
import numpy

from .. import geoprocessing
from . import utils


# TOLERANCE is based on the machine epsilon for single-precision floating-point
# decimal numbers.  See https://en.wikipedia.org/wiki/Machine_epsilon for more
# information on machine epsilons.  This is the default allowable relative
# error due to rounding for our assertions.
REL_TOL_DEFAULT = 1e-09
ABS_TOL_DEFAULT = 0.0
LOGGER = logging.getLogger('pygeoprocessing.testing.assertions')


def isclose(val_a, val_b, rel_tol=REL_TOL_DEFAULT, abs_tol=ABS_TOL_DEFAULT):
    """Assert that values are equal to the given tolerance.

    Adapted from the python 3.5 standard library based on the
    specification found in PEP485.

    Parameters:
        val_a (int or float): The first value to test
        val_b (int or float): The second value to test
        rel_tol (int or float): is the relative tolerance - it is the
            maximum allowed difference between a and b, relative to the
            larger absolute value of a or b. For example, to set a
            tolerance of 5%, pass `rel_tol=0.05`. The default tolerance
            is 1e-09, which assures that the two values are the same
            within about 9 decimal digits. rel_tol must be greater than
            zero.
        abs_tol (float): is the minimum absolute tolerance - useful for
            comparisons near zero. abs_tol must be at least zero.

    Returns:
        A boolean.
    """
    return abs(val_a-val_b) <= max(
        rel_tol * max(abs(val_a), abs(val_b)), abs_tol)


def assert_close(
        value_a, value_b, rel_tol=REL_TOL_DEFAULT, abs_tol=ABS_TOL_DEFAULT,
        msg=None):
    """
    Assert equality to an absolute tolerance.

    Parameters:
        value_a (int or float): The first value to test.
        value_b (int or float): The second value to test.
        rel_tol (int or float): The relative numerical tolerance.  If
            the relative difference of these values are less than this value,
            assertion will pass.
        abs_tol (int or float): The absolute numerical tolerance.  If the
            difference of the values being tested are less than this value,
            the assertion will pass.
        msg=None (string or None): The assertion message to use if value_a
            and value_b are not found to be equal out to the target tolerance.

    Returns:
        None.

    Raises:
        AssertionError: Raised when the values are not equal out to the
        desired tolerance.
    """
    if not isclose(value_a, value_b, rel_tol=rel_tol, abs_tol=abs_tol):
        if msg is None:
            msg = (
                "{a} != {b} within rel_tol {rel_tol} or "
                "abs_tol {abs_tol}".format(
                    a=value_a, b=value_b, rel_tol=rel_tol, abs_tol=abs_tol))
        raise AssertionError(msg)


def assert_rasters_equal(
        a_uri, b_uri, rel_tol=REL_TOL_DEFAULT, abs_tol=ABS_TOL_DEFAULT):
    """Assert te equality of rasters a and b out to the given tolerance.

    This assertion method asserts the equality of these raster
    characteristics:
        + Raster height and width

        + The number of layers in the raster

        + Each pixel value, such that the absolute difference between the
            pixels is less than `tolerance`.

        + Projection

    Args:
        a_uri (string): a URI to a GDAL dataset
        b_uri (string): a URI to a GDAL dataset
        rel_tol (int or float): the relative tolerance to which
            values should be asserted.  This is a numerical tolerance,
            not the number of places to which a value should be rounded.  If
            the relative value is below this value then the assert passes.
        abs_tol (int or float): absolute tolerance to which values should be
            asserted.  If absolute difference in values is below this value
            then the assert passes.

    Returns:
        None

    Raises:
        IOError: Raised when one of the input files is not found on disk.

        AssertionError: Raised when the two rasters are found to be not
        equal to each other.

    """
    for uri in [a_uri, b_uri]:
        if not os.path.exists(uri):
            raise IOError('File "%s" not found on disk' % uri)

    a_dataset = gdal.Open(a_uri)
    b_dataset = gdal.Open(b_uri)

    try:
        if a_dataset.RasterXSize != b_dataset.RasterXSize:
            raise AssertionError(
                "x dimensions are different a=%s, b=%s" %
                (a_dataset.RasterXSize, b_dataset.RasterXSize))

        if a_dataset.RasterYSize != b_dataset.RasterYSize:
            raise AssertionError(
                "y dimensions are different a=%s, b=%s" %
                (a_dataset.RasterYSize, b_dataset.RasterYSize))

        if a_dataset.RasterCount != b_dataset.RasterCount:
            raise AssertionError(
                "different number of rasters a=%s, b=%s" %
                ((a_dataset.RasterCount, b_dataset.RasterCount)))

        a_sr = osr.SpatialReference()
        a_sr.ImportFromWkt(a_dataset.GetProjection())

        b_sr = osr.SpatialReference()
        b_sr.ImportFromWkt(b_dataset.GetProjection())

        if bool(a_sr.IsSame(b_sr)) is False:
            raise AssertionError('Projections differ: %s != %s' % (
                a_sr.ExportToPrettyWkt(), b_sr.ExportToPrettyWkt()))

        for band_number in range(1, a_dataset.RasterCount + 1):
            a_band = a_dataset.GetRasterBand(band_number)
            b_band = b_dataset.GetRasterBand(band_number)

            a_blocksize = a_band.GetBlockSize()
            b_blocksize = b_band.GetBlockSize()
            if a_blocksize != b_blocksize:
                raise AssertionError(
                    'Block sizes differ for band %s: %s != %s' % (
                        band_number, a_blocksize, b_blocksize))

            for (_, a_block), (_, b_block) in zip(
                    geoprocessing.iterblocks(a_uri, [band_number]),
                    geoprocessing.iterblocks(b_uri, [band_number])):
                try:
                    numpy.testing.assert_allclose(
                        a_block, b_block, rtol=rel_tol, atol=abs_tol)
                except AssertionError:
                    iterator = numpy.nditer([a_block, b_block],
                                            flags=['multi_index'],
                                            op_flags=['readonly'])
                    while not iterator.finished:
                        col = iterator.multi_index[0]
                        row = iterator.multi_index[1]

                        pixel_a = a_block[iterator.multi_index]
                        pixel_b = b_block[iterator.multi_index]
                        assert_close(
                            pixel_a, pixel_b, rel_tol=rel_tol,
                            abs_tol=abs_tol, msg=(
                                '{a_val} != {b_val} at col {col}, '
                                'row {row} within rel_tol {rel_tol} and '
                                'abs_tol {abs_tol}').format(
                                    a_val=pixel_a, b_val=pixel_b, col=col,
                                    row=row, rel_tol=rel_tol,
                                    abs_tol=abs_tol))
                        iterator.iternext()
    finally:
        gdal.Dataset.__swig_destroy__(a_dataset)
        gdal.Dataset.__swig_destroy__(b_dataset)
        a_dataset = None
        b_dataset = None


def assert_vectors_equal(a_uri, b_uri, field_tolerance):
    """Assert that the vectors at a_uri and b_uri are equal to each other.

    This assertion method asserts the equality of these vector
    characteristics:
        + Number of layers in the vector

        + Number of features in each layer

        + Geometry type of the layer

        + Feature geometry type

        + Number of fields in each feature

        + Name of each field

        + Field values for each feature

        + Projection

    Args:
        a_uri (string): a URI to an OGR vector
        b_uri (string): a URI to an OGR vector
        field_tolerance (int or float): The relative numerical tolerance to
            which field values should be asserted.

    Raises:
        IOError: Raised if one of the input files is not found on disk.
        AssertionError: Raised if the vectors are not found to be equal to
        one another.

    Returns
        None
    """
    for uri in [a_uri, b_uri]:
        if not os.path.exists(uri):
            raise IOError('File "%s" not found on disk' % uri)

    shape = ogr.Open(a_uri)
    shape_regression = ogr.Open(b_uri)

    try:
        # Check that the shapefiles have the same number of layers
        layer_count = shape.GetLayerCount()
        layer_count_regression = shape_regression.GetLayerCount()
        if layer_count != layer_count_regression:
            raise AssertionError(
                'Number of vector layers do not match: %s != %s' % (
                    layer_count, layer_count_regression))

        for layer_num in range(layer_count):
            # Get the current layer
            layer = shape.GetLayer(layer_num)
            layer_regression = shape_regression.GetLayer(layer_num)
            # Check that each layer has the same number of features
            feat_count = layer.GetFeatureCount()
            feat_count_regression = layer_regression.GetFeatureCount()
            if feat_count != feat_count_regression:
                raise AssertionError(
                    ('The layers DO NOT have the same number of features'
                     '%s != %s' % (feat_count, feat_count_regression)))

            layer_geom = layer.GetGeomType()
            layer_reg_geom = layer_regression.GetGeomType()
            if layer_geom != layer_reg_geom:

                def _readable_geometry_type(geom_type):
                    """Determine the geometry type from the OGR wkb int.

                    Parameters:
                        geom_type: The result of ogr.Layer.GetGeomType()

                    Returns:
                        A string representation of the feature type.
                    """
                    for attr_name in dir(ogr):
                        if (attr_name.startswith('wkb') and
                                getattr(ogr, attr_name) == geom_type):
                            return attr_name.replace('wkb', '')

                layer_geom_string = _readable_geometry_type(layer_geom)
                layer_reg_geom_string = _readable_geometry_type(layer_reg_geom)
                raise AssertionError(
                    'Layers #%s have different geometry types: %s != %s' % (
                        layer_num, layer_geom_string, layer_reg_geom_string))

            a_sr = layer.GetSpatialRef()
            b_sr = layer_regression.GetSpatialRef()
            if bool(a_sr.IsSame(b_sr)) is False:
                raise AssertionError(
                    'Projections differ in layer %s: %s != %s' % (
                        layer_num, a_sr.ExportToPrettyWkt(),
                        b_sr.ExportToPrettyWkt()))

            # Get the first features of the layers and loop through all the
            # features
            feat = layer.GetNextFeature()
            feat_regression = layer_regression.GetNextFeature()
            while feat is not None:
                # Check that the field counts for the features are the same
                layer_def = layer.GetLayerDefn()
                layer_def_regression = layer_regression.GetLayerDefn()
                field_count = layer_def.GetFieldCount()
                field_count_regression = layer_def_regression.GetFieldCount()
                if field_count != field_count_regression:
                    raise AssertionError((
                        'The features DO NOT have the same number of fields '
                        'in layer %s: %s != %s') % (
                            layer_num, field_count, field_count_regression))

                for fld_index in range(field_count):
                    # Check that the features have the same field values
                    field = feat.GetField(fld_index)
                    field_regression = feat_regression.GetField(fld_index)

                    try:
                        try:
                            float_field = float(field)
                            float_field_regression = float(field_regression)
                            assert_close(
                                float_field, float_field_regression,
                                field_tolerance)
                        except (ValueError, TypeError):
                            # ValueError happens when strings cast to float
                            # TypeError when casting non-string or non-number.
                            if field != field_regression:
                                raise AssertionError
                    except AssertionError:
                        # If we found an AssertionError, raise a more helpful
                        # assertion error mesage here that includes vector
                        # information.
                        raise AssertionError(
                            'Field values %s != %s at index %s in layer %s' % (
                                field, field_regression, fld_index, layer_num))

                    # Check that the features have the same field name
                    field_ref = feat.GetFieldDefnRef(fld_index)
                    field_ref_regression = \
                        feat_regression.GetFieldDefnRef(fld_index)
                    field_name = field_ref.GetNameRef()
                    field_name_regression = field_ref_regression.GetNameRef()
                    if field_name != field_name_regression:
                        raise AssertionError(
                            'Field names %s != %s at index %s in layer %s' % (
                                field_name, field_name_regression, fld_index,
                                layer_num))

                # Check that the features have the same geometry
                geom = feat.GetGeometryRef()
                geom_regression = feat_regression.GetGeometryRef()

                if bool(geom.Equals(geom_regression)) is False:
                    feature_fid = feat.GetFID()
                    reg_feature_fid = feat_regression.GetFID()
                    raise AssertionError(
                        'Geometries are not equal in feature %s, '
                        'regression feature %s in layer %s' % (
                            feature_fid, reg_feature_fid, layer_num))

                feat = None
                feat_regression = None
                feat = layer.GetNextFeature()
                feat_regression = layer_regression.GetNextFeature()
    finally:
        ogr.DataSource.__swig_destroy__(shape)
        ogr.DataSource.__swig_destroy__(shape_regression)
        shape = None
        shape_regression = None


def assert_csv_equal(
        a_uri, b_uri, rel_tol=REL_TOL_DEFAULT, abs_tol=ABS_TOL_DEFAULT):
    """Assert the equality of CSV files at a_uri and b_uri.

    Tests if csv files a and b are 'almost equal' to each other on a per
    cell basis.  Numeric cells are asserted to be equal within the given
    tolerance.  Other cell types are asserted to be equal.

    Args:
        a_uri (string): a URI to a csv file
        b_uri (string): a URI to a csv file
        rel_tol (int or float): The relative numerical tolerance allowed, if
            relative difference of values are less than this, assertion
            passes.
        abs_tol (int or float): The absolute numerical tolerance allowed, if
            difference of values are less than this, assertion passes.

    Raises:
        AssertionError: Raised when the two CSV files are found to be
        different.

    Returns:
        None
    """
    a_file = open(a_uri, 'r')
    b_file = open(b_uri, 'r')

    reader_a = csv.reader(a_file)
    reader_b = csv.reader(b_file)

    for index, (a_row, b_row) in enumerate(zip(reader_a, reader_b)):
        try:
            if a_row != b_row:
                raise AssertionError('Rows differ at row'
                                     '%s: a=%s b=%s' % (index, a_row, b_row))
        except AssertionError:
            for col_index, (a_element, b_element) in enumerate(zip(a_row,
                                                                   b_row)):
                try:
                    a_element = float(a_element)
                    b_element = float(b_element)
                    assert_close(
                        a_element, b_element, rel_tol=rel_tol,
                        abs_tol=abs_tol, msg=(
                            'Values are significantly different at row %s'
                            'col %s: a=%s b=%s' % (
                                index, col_index, a_element, b_element)))
                except ValueError:
                    # we know for sure they arenot floats, so compare as
                    # non-floats.
                    if a_element != b_element:
                        raise AssertionError(
                            'Elements differ at row %s col%s: a=%s '
                            'b=%s' % (index, col_index, a_element, b_element))


def assert_md5_equal(uri, regression_hash):
    """Assert the MD5sum of a file against a regression MD5sum.

    This method is a convenience method that uses
    ``natcap.invest.testing.digest_file()`` to determine the MD5sum of the
    file located at `uri`.  It is functionally equivalent to calling::

        if digest_file(uri) != '<some md5sum>':
            raise AssertionError

    Regression MD5sums can be calculated for you by using
    ``natcap.invest.testing.digest_file()`` or a system-level md5sum program.

    Args:
        uri (string): a string URI to the file to be tested.
        regression_hash (string): a string md5sum to be tested against.

    Raises:
        AssertionError: Raised when the MD5sum of  the file at `uri`
        differs from the provided regression md5sum hash.

    Returns:
        None
    """
    digested_file = utils.digest_file(uri)
    if digested_file != regression_hash:
        raise AssertionError('MD5 hashes differ: %s != %s' % (digested_file,
                                                              regression_hash))


def assert_json_equal(json_1_uri, json_2_uri):
    """Assert two JSON files against each other.

    The two JSON files provided will be opened, read, and their
    contents will be asserted to be equal.  If the two are found to be
    different, the diff of the two files will be printed.

    Args:
        json_1_uri (string): a uri to a JSON file.
        json_2_uri (string): a uri to a JSON file.

    Raises:
        AssertionError: Raised when the two JSON objects differ.

    Returns:
        None
    """
    dict_1 = json.loads(open(json_1_uri).read())
    dict_2 = json.loads(open(json_2_uri).read())

    if dict_1 != dict_2:
        raise AssertionError('JSON objects differ: %s\n%s' % (dict_1, dict_2))


def assert_text_equal(text_1_uri, text_2_uri):
    """Assert that two text files are equal.

    This comparison is done line-by-line.

    Args:
        text_1_uri (string): a python string uri to a text file.
            Considered the file to be tested.
        text_2_uri (string): a python string uri to a text file.
            Considered the regression file.

    Raises:
        AssertionError: Raised when a line differs in the two files.

    Returns:
        None
    """
    def lines(filepath):
        """Return a generator of lines in the opened file."""
        with open(filepath, 'rb') as opened_file:
            for line in opened_file:
                yield line

    for index, (a_line, b_line) in enumerate(zip(
            lines(text_1_uri), lines(text_2_uri))):
        if a_line != b_line:
            raise AssertionError('Line %s in %s does not match regression '
                                 'file %s. Output  "%s" Regression "%s"' % (
                                     index, text_1_uri, text_2_uri, a_line,
                                     b_line))


def assert_checksums_equal(checksum_file, base_folder=None):
    """Assert all files in the `checksum_file` have the same checksum.

    Checksum files could be created by
    `pygeoprocessing.testing.utils.checksum_folder()`, but this function
    may also assert the output of GNU `md5sum` or BSD `md5`.  Either format
    (GNU or BSD) may be provided as input to this assertion function.
    Any files not included in `checksum_file` are ignored.

    Parameters:
        checksum_file (string): the path to the snapshot file to use.
        base_folder=None (string or None): the folder to refer to as the base
            path for files in the `checksum_file`.  If `None`, the current
            working directory will be used.

    Raises:
        AssertionError: when a nonmatching md5sum is found.
    """
    if base_folder is None:
        base_folder = os.getcwd()

    snapshot = open(checksum_file)
    env_params = {}
    files = {}
    for line in snapshot:
        # environment parameters are recorded as comments.
        if line.startswith('#'):
            name, value = line.split('=')
            name = name.replace('#', '').strip()
            env_params[name.strip()] = value.strip()
        else:
            split_line = line.strip().split(' ')
            if split_line[0] == 'MD5':
                # we're reading a BSD-style checksum file with format:
                # MD5 (filename) = md5sum
                md5sum = split_line[3]
                filename = split_line[1].replace(')', '').replace('(', '')
            else:
                # We're reading a GNU-style checksum file with format:
                # md5sum  filename
                md5sum = split_line[0]
                filename = split_line[2]

            files[filename] = md5sum

    missing_files = []
    nonmatching_files = []
    for filepath, expected_md5sum in files.items():
        full_filepath = os.path.join(base_folder, filepath)
        try:
            current_md5sum = utils.digest_file(full_filepath)
        except IOError:
            # When the file we're looking for doesn't exist
            # Keep around for reporting missing files later.
            missing_files.append(full_filepath)
            continue

        if current_md5sum != expected_md5sum:
            nonmatching_files.append(filepath)

    if len(missing_files) != 0:
        if len(missing_files) == len(files):
            raise AssertionError((
                'All files recorded in the snapshot are missing.  Are you '
                'testing against the right folder?  Testing {test_dir}. '
                'Snapshot taken from {snap_dir}.').format(
                    test_dir=base_folder,
                    snap_dir=env_params['orig_workspace']))
        raise AssertionError(
            ('{num_missing} files out of {num_files} are '
             'missing.').format(num_missing=len(missing_files),
                                num_files=len(files)))

    if len(nonmatching_files) != 0:
        raise AssertionError((
            '{num_err} files out of {num_files} have differing '
            'md5sums: {files}').format(num_err=len(nonmatching_files),
                                       num_files=len(files),
                                       files=nonmatching_files))
